#!/usr/bin/env python
# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
#
# ---------------------------------------------------------


from bokeh.plotting import output_file

# Standard title attribute for all figures
title_style = {'title_text_font_size': '12pt',
               'title_text_font_style': 'bold'}

grid_title_style = {'title_text_font_size': '10pt',
                    'title_text_font_style': 'bold'}

def set_html_file(cdict_file):
    global html_file
    html_file = cdict_file.replace('.cdict', '')

# calculate the time between the 1st entry and the last entry in msec
def get_time_span_msec(df):
    min_usec = df['usecs'].min()
    max_usec = df['usecs'].max()
    return (max_usec - min_usec) / 1000

# For sorting
# 'CSR.1.vcpu0' => '0001.CSR.vcpu0'
def normalize_task_name(task):
    tokens = task.split('.')
    if len(tokens) > 1:
        # extract the service chain
        chain = tokens[1]
        try:
            chain = int(tokens[1])
            res = '%02d' % (chain)
            tokens.pop(1)
            return res + '.' + '.'.join(tokens)
        except ValueError:
            pass
    return task

def normalize_df_task_name(df):
    df['task_name'] = df.apply(lambda row: normalize_task_name(row['task_name']), axis=1)

def get_disc_size(count):
    if count < 100:
        return 10
    if count < 500:
        return 8
    return 6

def output_html(chart_type, task_re):
    filename = html_file + '_' + chart_type + '_' + task_re + '.html'
    output_file(filename)
    print('Saved to ' + filename)

def get_full_task_name(df, task):
    # if task is a number it is considered to be a pid ID
    # if text it is a task name
    try:
        tid = int(task)
        # tid given
        df = df[df['pid'] == tid]
        task = None
        # get the task name from the tid
    except ValueError:
        # task given: find corresponding tid
        df = df[df['task_name'] == task]
        tid = 0

    if df.empty:
        print 'No selection matching the task ' + task
        return (None, None)
    # fill in the missing information
    if not tid:
        tid = df['pid'].iloc[0]
    if not task:
        task = df['task_name'].iloc[0]
    task = task + ':' + str(tid)
    return (df, task)



#!/usr/bin/env python
# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
#
# Author: Alec Hothan
# ---------------------------------------------------------


from collections import OrderedDict
import pandas
from pandas import Series

from bokeh.plotting import show
from bokeh.models.sources import ColumnDataSource
from bokeh.models import HoverTool
from bokeh.io import vplot
from bokeh.charts import Bar
from bokeh.models.widgets import DataTable
from bokeh.models.widgets import TableColumn
from bokeh.models.widgets import NumberFormatter
from bokeh.models.widgets import StringFormatter

from perfmap_common import output_html
from perfmap_common import get_time_span_msec

# KVM exit reasons
# Intel64 and IA32 Architecture Software Developer's Manual Vol 3B, System Programming Guide Part 2
# Appendix I
KVM_EXIT_REASONS = [
    'Exception or NMI',     # 0
    'External Interrupt',
    'Triple Fault',
    'INIT',
    'Startup IPI',
    'I/O SMI (System Management Interrupt)',
    'Other SMI',
    'Interrupt Window',
    'NMI window',
    'Task Switch',
    'CPUID',                # 10
    'GETSEC',
    'HLT',                  # 12
    'INVD',
    'INVLPG',
    'RDPMC',
    'RDTSC',
    'RSM',
    'VMCALL',
    'VMCLEAR',
    'VMLAUNCH',             # 20
    'VMPTRLD',
    'VMPTRST',
    'VMREAD',
    'VMRESUME',
    'VMWRITE',
    'VMXOFF',
    'VMXON',
    'Control Register Access',
    'MOV DR',
    'I/O Instruction',      # 30
    'RDMSR',
    'WRMSR',
    'VM Entry Failure (invalid guest state)',
    'VM Entry Failure (MSR loading)',
    'n/a 35',
    'MWAIT',
    'Monitor trap flag',
    'n/a 38',
    'MONITOR',
    'PAUSE',                # 40
    'VM Entry Failure (machine check)',
    'n/a 42',
    'TPR below threshold',
    'APIC Access',
    'n/a 45',
    'Access to GDTR or IDTR',
    'Access to LDTR or TR',
    'EPT violation',
    'EPT misconfiguration',
    'INVEPT',               # 50
    'RDTSCP',
    'VMX preemption timer expired',  # 52
    'INVVPID',
    'WBINVD',
    'XSETBV'
]

def convert_exit_df(df, label):
    # fill in the error reason text from the code in
    df.next_comm = df.next_comm.apply(lambda x: '(%02d) %s' % (x, KVM_EXIT_REASONS[x]))
    series_percent = df.next_comm.value_counts(normalize=True)
    series_count = df.next_comm.value_counts()
    series_percent = pandas.Series(["{0:.2f}%".format(val * 100) for val in series_percent],
                                   index=series_percent.index)

    series_percent.name = '%'
    series_count.name = label + ' count'

    res = pandas.concat([series_count, series_percent], axis=1)
    return res

def show_exit_type_count(df_all_exits, df_last_exits):

    res_all = convert_exit_df(df_all_exits, 'total')
    res_last = convert_exit_df(df_last_exits, 'last exit')
    res = pandas.concat([res_all, res_last], axis=1)
    res.fillna(0, inplace=True)
    res.sort_values('total count', inplace=True, ascending=False)
    print res

def show_kvm_exit_types(df, task_re, label):

    df = df[df['event'] == 'kvm_exit']
    df = df[df['task_name'].str.match(task_re)]
    # the next_comm column contains the exit code
    exit_codes = Series(KVM_EXIT_REASONS)
    # add  new column congaining the exit reason in clear text
    df['exit_reason'] = df['next_comm'].map(exit_codes)
    time_span_msec = get_time_span_msec(df)
    df.drop(['cpu', 'duration', 'event', 'next_pid', 'pid', 'next_comm', 'usecs'], inplace=True, axis=1)

    # Get the list of exit reasons, sorted alphabetically
    reasons = pandas.unique(df.exit_reason.ravel()).tolist()
    reasons.sort()

    # group by task name then exit reasons
    gb = df.groupby(['task_name', 'exit_reason'])
    # number of exit types
    size_series = gb.size()
    df = size_series.to_frame('count')
    df.reset_index(inplace=True)

    p = Bar(df, label='task_name', values='count', stack='exit_reason',
            title="KVM Exit types per task (%s, %d msec window)" % (label, time_span_msec),
            legend='top_right',
            tools="resize,hover,save",
            width=1000, height=800)
    p._xaxis.axis_label = "Task Name"
    p._xaxis.axis_label_text_font_size = "12pt"
    p._yaxis.axis_label = "Exit Count (sum)"
    p._yaxis.axis_label_text_font_size = "12pt"

    # Cannot find a way to display the exit reason in the tooltip
    # from bokeh.models.renderers import GlyphRenderer
    # glr = p.select(dict(type=GlyphRenderer))
    # bar_source = glr[0].data_source
    # print bar_source.data
    # bar_source = glr[1].data_source
    # bar_source.data['exit_reason'] = ['HOHO']
    hover = p.select(dict(type=HoverTool))
    hover.tooltips = OrderedDict([
        ("task", "$x"),
        # {"reason", "@exit_reason"},
        ("count", "@height")
    ])
    # specify how to output the plot(s)
    output_html('kvm-types', task_re)

    # table with counts
    gb = df.groupby(['exit_reason'])
    keys = gb.groups.keys()
    dfr_list = []
    for reason in keys:
        dfr = gb.get_group(reason)
        # drop the exit reason column
        dfr = dfr.drop(['exit_reason'], axis=1)
        # rename the count column with the reason name
        dfr.rename(columns={'count': reason}, inplace=True)
        # set the task name as the index
        dfr.set_index('task_name', inplace=True)
        dfr_list.append(dfr)
    # concatenate all task columns into 1 dataframe that has the exit reason as the index
    # counts for missing exit reasons will be set to NaN
    dft = pandas.concat(dfr_list, axis=1)
    dft.fillna(0, inplace=True)
    # Add a total column
    dft['TOTAL'] = dft.sum(axis=1)
    sfmt = StringFormatter(text_align='center', font_style='bold')
    nfmt = NumberFormatter(format='0,0')

    col_names = list(dft.columns.values)
    col_names.sort()
    # move 'TOTAL' at end of list
    col_names.remove('TOTAL')
    col_names.append('TOTAL')
    # convert index to column name
    dft.reset_index(level=0, inplace=True)
    dft.rename(columns={'index': 'Task'}, inplace=True)
    columns = [TableColumn(field=name, title=name, formatter=nfmt) for name in col_names]
    columns.insert(0, TableColumn(field='Task', title='Task', formatter=sfmt))
    table = DataTable(source=ColumnDataSource(dft), columns=columns, width=1000,
                      row_headers=False,
                      height='auto')
    show(vplot(p, table))

    '''
    # Show aggregated time inside VM and inside KVM
    print
    print 'Aggregated duration:'
    for event in legend_map:
        percent = accumulated_time[event] * 100 / total_time
        print '   %40s: %9d usec %3d%%' % (legend_map[event][1], accumulated_time[event], percent)
    print '   %40s: %9d usec %3d%%' % ('Total', total_time, 100)
    print

    # all exit types
    dfexits = df[df.event == 'kvm_exit']

    # now keep only the last exit before a switch
    # only keep kvm_exits and context switches
    df = df[(df['event'] == 'kvm_exit') | (df['event'] == 'sched__sched_switch')]

    # move all rows up
    sdf = df.shift(-1)
    # only keep rows that have a different event than their next row
    df = df[df['event'] != sdf['event']]
    # delete last row
    df = df.drop(df.tail(1).index)

    # delete all sched switches as they have served their purpose
    df = df[df['event'] == 'kvm_exit']

    show_exit_type_count(dfexits, df)
    '''



#!/usr/bin/env python
# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
#
# Author: Alec Hothan
# ---------------------------------------------------------


from collections import OrderedDict
import itertools
import pandas
from pandas import DataFrame
import numpy as np

from bokeh.plotting import figure, show
from bokeh.models.sources import ColumnDataSource
from bokeh.models import HoverTool
from bokeh.palettes import Spectral6
from bokeh.palettes import YlOrRd9

from perfmap_common import title_style
from perfmap_common import output_html
from perfmap_common import get_time_span_msec
from perfmap_common import get_disc_size

# For sorting
# 'CSR.1.vcpu0' => '0001.CSR.vcpu0'
def normalize_task_name(task):
    tokens = task.split('.')
    if len(tokens) > 1:
        # extract the service chain
        chain = tokens[1]
        try:
            chain = int(tokens[1])
            res = '%02d' % (chain)
            tokens.pop(1)
            return res + '.' + '.'.join(tokens)
        except ValueError:
            pass
    return task

def normalize_df_task_name(df):
    df['task_name'] = df.apply(lambda row: normalize_task_name(row['task_name']), axis=1)

GRAY = "#b5cbc5"

def cycle_colors(chunk, palette=Spectral6):
    """ Build a color list just cycling through a given palette.

    Args:
        chuck (seq): the chunk of elements to generate the color list
        palette (seq[color]) : a palette of colors to cycle through

    Returns:
        colors

    """
    colors = []

    g = itertools.cycle(palette)
    for i in range(len(chunk)):
        colors.append(next(g))

    return colors

def get_color(percent, palette):
    try:
        percent = int(percent)
    except ValueError:
        # swapper tasks always have NaN since their duration is always 0
        percent = 100
    if percent >= 100:
        return palette[-1]
    return palette[percent * len(palette) // 100]

def get_color_value_list(min_count, max_count, palette, range_unit):
    value_list = []
    color_count = len(palette)
    increment = float(max_count - min_count) / color_count
    value_list = np.arange(min_count, max_count + 1, increment).astype(np.int).astype(str).tolist()
    if range_unit:
        value_list = [x + range_unit for x in value_list]
    return value_list

def show_core_runs(df, task_re, label, duration):
    time_span_msec = get_time_span_msec(df)

    # remove unneeded columns
    df = df.drop('next_pid', axis=1)
    df.drop('pid', axis=1, inplace=True)
    df.drop('usecs', axis=1, inplace=True)
    df.drop('next_comm', axis=1, inplace=True)

    # filter out all events except the switch events
    df = df[df.event == 'sched__sched_switch']
    df = df.drop('event', axis=1)
    df = df[df['task_name'].str.match(task_re)]

    # at this point we have a df that looks like this:
    #         task_name  cpu  duration
    # 0     ASA.1.vcpu0    8      7954
    # 1     ASA.1.vcpu0    9      5475
    # 2     ASA.1.vcpu0   10      4151
    # 3     ASA.1.vcpu0   11     12391
    # 4     ASA.1.vcpu0   12     21025
    # 5     ASA.1.vcpu0   13      6447
    # 6     ASA.1.vcpu0   14     16798
    # 7     ASA.1.vcpu0   15      3911
    # 8    ASA.10.vcpu0    8      4248
    # 9    ASA.10.vcpu0    9      3534
    # 10   ASA.10.vcpu0   10     15624
    # 11   ASA.10.vcpu0   11      6925
    # etc...
    if len(df) == 0:
        print
        print 'No selection matching "%s"' % (task_re)
        return
    gb = df.groupby(['task_name', 'cpu'], as_index=False)
    if duration:
        # add duration values
        df = gb.aggregate(np.sum)
        max_core = df.cpu.max()

        dfsum = df.drop('cpu', axis=1)
        gb = dfsum.groupby('task_name', as_index=False)
        dfsum = gb.aggregate(np.sum)
        # dfsum is the sum of all duration for given task
        # 0    ASA.1.vcpu0     78152
        # 1   ASA.10.vcpu0     65637
        # 2   ASA.11.vcpu0     81525
        # 3   ASA.12.vcpu0     56488
        # For each task, the maximum runtime is the time_span_msec (100% of 1 core)
        # The idle time for each task is therefore time_span_msec * 1000 - sum(duration)
        dfsum['cpu'] = 'IDLE'
        time_span_usec = time_span_msec * 1000
        dfsum['duration'] = time_span_usec - dfsum['duration']

        # now we need to reinsert that data back to the df
        dfm = pandas.concat([df, dfsum], ignore_index=True)

        #         task_name  cpu  duration  total
        # 0     ASA.1.vcpu0    8      7954  78152
        # 1     ASA.1.vcpu0    9      5475  78152
        # 2     ASA.1.vcpu0   10      4151  78152

        # Add a % column
        dfm['percent'] = ((dfm['duration'] * 100 * 10) // time_span_usec) / 10

        # This is for the legend
        min_count = 0
        max_count = 100
        range_unit = '%'

        # many core-pinned system tasks have a duration of 0 (swapper, watchdog...)
        dfm.fillna(100, inplace=True)
        dfm.drop(['duration'], axis=1, inplace=True)
        tooltip_count = ("time", "@percent% of core @cpu")
        title = "Task Run Time %% per Core (%s, %d msec window)" % (label, time_span_msec)
        # this is YlGnBu9[::-1]  (Reverse the color order so dark is highest value)
        # with an extra intermediate color to make it 10
        palette = ['#ffffd9', '#edf8b1', '#c7e9b4', '#a3dbb7', '#7fcdbb',
                   '#41b6c4', '#1d91c0', '#225ea8', '#253494', '#081d58']
        html_prefix = 'core_runtime'
        # add 1 extra column in the heatmap for core "IDLE" to represent the IDLE time
        # this requires adding 1 row per real core that has a runtime set to
        # capture window size - sum(duration on all cores)
    else:
        # count number of rows with same task and cpu
        dfm = DataFrame(gb.size())
        dfm.reset_index(inplace=True)
        dfm.rename(columns={0: 'count'}, inplace=True)
        min_count = dfm['count'].min()
        max_count = dfm['count'].max()
        range_unit = ''
        spread = max_count - min_count
        # Add a % column
        dfm['percent'] = ((dfm['count'] - min_count) * 100) / spread
        tooltip_count = ("context switches", "@count")
        title = "Task Context Switches per Core (%s, %d msec window)" % (label, time_span_msec)
        palette = YlOrRd9[::-1]
        html_prefix = 'core_switches'
        max_core = dfm.cpu.max()

    # get the list of cores
    # round up to next mutliple of 4 - 1
    # 0..3 -> 3
    # 4..7 -> 7 etc
    max_core |= 0x03
    max_core = max(max_core, 3)
    core_list = [str(x) for x in range(max_core + 1)]
    if duration:
        core_list.append('IDLE')
    # make room for the legend by adding 3 empty columns
    core_list += ['', '', '']
    dfm['cpu'] = dfm['cpu'].astype(str)
    # replace ':' with '_' as it would cause bokeh to misplace the labels on the chart
    dfm['task_name'] = dfm['task_name'].str.replace(':', '_')

    normalize_df_task_name(dfm)

    # Add a column for the Y axis
    # each task name should be associated to a unique Y index
    # Get a unique list of task names sorted
    task_list = pandas.unique(dfm.task_name.ravel()).tolist()
    task_list.sort()

    # Add a color column
    dfm['color'] = dfm.apply(lambda row: get_color(row['percent'],
                                                   palette), axis=1)
    # switch to str type to prevent the tooltip to
    # display percent value with 3 digits
    dfm.percent = dfm.percent.astype(str)
    # make enough vertical space for the legend
    # the legend needs 1 row per palette color + 1 to fit the max value
    # so we need at least len(palette) + 1 rows
    if len(task_list) < len(palette) + 1:
        task_list += ['' for _ in range(len(palette) + 1 - len(task_list))]
    TOOLS = "resize,hover,save"
    p = figure(title=title, tools=TOOLS, x_range=core_list, y_range=task_list, **title_style)

    p.plot_width = 1000
    p.plot_height = 80 + len(task_list) * 16
    p.toolbar_location = "left"
    source = ColumnDataSource(dfm)
    # the name is to flag these rectangles to enable tooltip hover on them
    # (and not enable tooltips on the legend patches)
    p.rect("cpu", "task_name", width=1, height=0.9, source=source, fill_alpha=0.8, color="color", name='patches')
    p.grid.grid_line_color = None
    # trace separator lines to isolate blocks across core groups (numa sockets) and task-like names
    max_y = len(task_list)
    # trace a vertical line every 8 cores
    for seg_x in range(8, max_core + 7, 8):
        p.segment(x0=[seg_x + 0.5], y0=[0], x1=[seg_x + 0.5],
                  y1=[max_y + 0.5], color=GRAY, line_width=2)
    prev_task_name = None
    # trace a horizontal line around every group of tasks that have the same first 3 characters
    for y in range(max_y):
        cur_task_name = task_list[y]
        if prev_task_name:
            if prev_task_name[0:3] != cur_task_name[0:3]:
                p.segment(x0=[0], y0=[y + 0.5], x1=[max_core + 1.5],
                          y1=[y + 0.5], color=GRAY, line_width=0.5)
        prev_task_name = task_list[y]

    hover = p.select(dict(type=HoverTool))
    hover.tooltips = OrderedDict([
        ("task", "@task_name"),
        ("core", "@cpu"),
        tooltip_count
    ])
    # only enable tooltip on rectangles with name 'patches'
    hover.names = ['patches']

    # legend to the right
    # we try to center the legend vertically
    legend_base_y = 0
    palette_len = len(palette)
    if max_y > palette_len + 1:
        legend_base_y = (max_y - palette_len - 1) // 2

    if duration:
        # IDLE cpu inserted so shift the legend by 1 position to the right
        max_core += 1
    # pass 1 is to draw the color patches
    # prepare a data source with a x, y and a color column
    x_values = np.empty(palette_len)
    x_values.fill(max_core + 2.5)
    y_values = np.arange(legend_base_y + 1.5, legend_base_y + palette_len + 1, 1)
    dfl = DataFrame({'x': x_values, 'y': y_values, 'color': palette})
    source = ColumnDataSource(dfl)
    p.rect(x='x', y='y', color='color', width=1, height=1, source=source)

    # pass 2 is to draw the text describing the ranges for the color patches
    color_value_list = get_color_value_list(min_count, max_count, palette, range_unit)
    x_values = np.empty(len(color_value_list))
    x_values.fill(max_core + 3.1)
    y_values = np.arange(legend_base_y + 0.7, legend_base_y + len(color_value_list), 1)

    dfl = DataFrame({'x': x_values, 'y': y_values, 'color_values': color_value_list})
    source = ColumnDataSource(dfl)
    p.text(x='x', y='y', text='color_values', source=source,
           text_font_size='8pt')

    output_html(html_prefix, task_re)
    show(p)


def show_core_locality(df, task_re, label):

    df = df[df['event'] == 'sched__sched_switch']
    df = df[df['task_name'].str.match(task_re)]
    # aggregate all the per core tasks (e.g. swapper/0 -> swapper)
    df['task_name'] = df['task_name'].str.replace(r'/.*$', '')
    # group by task name
    gb = df.groupby('task_name')
    task_list = gb.groups

    p = figure(plot_width=1000, plot_height=800, **title_style)

    p.xaxis.axis_label = 'time (usecs)'
    p.yaxis.axis_label = 'core'
    p.legend.orientation = "bottom_right"
    p.xaxis.axis_label_text_font_size = "10pt"
    p.yaxis.axis_label_text_font_size = "10pt"
    p.title = "Core locality (%s)" % (label)

    color_list = cycle_colors(task_list)

    for task, color in zip(task_list, color_list):
        dfe = gb.get_group(task)
        # add 1 column to contain the starting time for each run period
        dfe['start'] = dfe['usecs'] - dfe['duration']
        tid = dfe['pid'].iloc[0]
        count = len(dfe)
        legend_text = '%s:%d (%d)' % (task, tid, count)
        # draw end of runs
        p.circle('usecs', 'cpu', source=ColumnDataSource(dfe),
                 size=get_disc_size(count) + 2, color=color,
                 alpha=0.3,
                 legend=legend_text)
        # draw segments to show the entire runs
        p.segment('start', 'cpu', 'usecs', 'cpu', line_width=5, line_color=color,
                  source=ColumnDataSource(dfe))

    # specify how to output the plot(s)
    output_html('coreloc', task_re)

    # display the figure
    show(p)



#!/usr/bin/env python
# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
#
# Author: Alec Hothan
# ---------------------------------------------------------


from optparse import OptionParser
import os
import sys
import warnings
import pandas
from pandas import DataFrame

from perf_formatter import open_cdict
from perf_formatter import write_cdict

from perfmap_common import set_html_file

from perfmap_kvm_exit_types import show_kvm_exit_types
from perfmap_sw_kvm_exits import show_sw_kvm_heatmap

from perfmap_core import show_core_runs
from perfmap_core import show_core_locality

# Global variables

# start analysis after first from_time usec
from_time = 0
# cap input file to first cap_time usec, 0 = unlimited
cap_time = 0

def get_full_task_name(df, task):
    # if task is a number it is considered to be a pid ID
    # if text it is a task name
    try:
        tid = int(task)
        # tid given
        df = df[df['pid'] == tid]
        task = None
        # get the task name from the tid
    except ValueError:
        # task given: find corresponding tid
        df = df[df['task_name'] == task]
        tid = 0

    if df.empty:
        print 'No selection matching the task ' + task
        return (None, None)
    # fill in the missing information
    if not tid:
        tid = df['pid'].iloc[0]
    if not task:
        task = df['task_name'].iloc[0]
    task = task + ':' + str(tid)
    return (df, task)

def show_successors(df, task, label):
    df, task = get_full_task_name(df, task)
    if not task:
        return
    df = df[df['event'] == 'sched__sched_switch']
    # aggregate all the per core tasks (e.g. swapper/0 -> swapper)
    df['next_comm'] = df['next_comm'].str.replace(r'/.*$', '')
    series_percent = df.next_comm.value_counts(normalize=True)
    series_count = df.next_comm.value_counts()
    series_percent = pandas.Series(["{0:.2f}%".format(val * 100) for val in series_percent],
                                   index=series_percent.index)

    series_percent.name = 'percent'
    series_count.name = 'count'
    print 'Successors of %s (%s)' % (task, label)
    print pandas.concat([series_count, series_percent], axis=1)

def convert(df, new_cdict):
    # Convert the old style cdict (marshal based) to the new cdict format
    # The new format is msgpack based and fixes the runtime reporting bug
    # by aggregating all runtime durations into the next switch event
    # and removing the runtime events
    # A dict of runtime delays accumulated indexed by cpu
    runtime_by_cpu = {}
    # aggregate all runtimes per cpu
    count = len(df)
    print 'Conversion in progress...'
    for index in xrange(count):
        event_name = df['event'].iloc[index]
        cpu = df['cpu'].iloc[index]
        if event_name == 'sched__sched_stat_runtime':
            try:
                duration = df['duration'].iloc[index]
                runtime_by_cpu[cpu] += duration
            except KeyError:
                # the counter is set in on the first sched switch for each cpu
                pass
        elif event_name == 'sched__sched_switch':
            try:
                if df['pid'].iloc[index]:
                    # don't bother to update swapper (pid=0)
                    runtime = runtime_by_cpu[cpu]
                    df.set_value(index, 'duration', runtime)
            except KeyError:
                pass
            runtime_by_cpu[cpu] = 0
    # get rid of all the runtime events
    df = df[df['event'] != 'sched__sched_stat_runtime']
    print 'End of conversion, marshaling and compressing...'
    df.fillna(value=0, inplace=True, downcast='infer')
    # save new cdict
    res = {'event': df['event'].tolist(),
           'cpu': df['cpu'].tolist(),
           'usecs': df['usecs'].tolist(),
           'pid': df['pid'].tolist(),
           'task_name': df['task_name'].tolist(),
           'duration': df['duration'].tolist(),
           'next_pid': df['next_pid'].tolist(),
           'next_comm': df['next_comm'].tolist()}
    write_cdict(new_cdict, res)

# ---------------------------------- MAIN -----------------------------------------
# Suppress future warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

parser = OptionParser(usage="usage: %prog [options] <cdict_file>")


parser.add_option("--successors-of",
                  dest="successor_of_task",
                  help="show list of successors of given tid or task name"
                  )
parser.add_option("--show-tids",
                  dest="show_tids",
                  action="store_true",
                  default=False,
                  help="show list of all tids with event count)"
                  )
parser.add_option("-t", "--task",
                  dest="task",
                  metavar="task name (regex)",
                  help="selected task(s) (regex on task name)"
                  )
parser.add_option("--core-locality",
                  dest="core_loc",
                  action="store_true",
                  help="show core locality heat map (requires --task)"
                  )
parser.add_option("--core-runtime",
                  dest="core_runtime",
                  action="store_true",
                  help="show % runtime on each core (requires --task)"
                  )
parser.add_option("--core-switch-count",
                  dest="core_switches",
                  action="store_true",
                  help="show context switch count on each core (requires --task)"
                  )
parser.add_option("--switches",
                  dest="switches",
                  action="store_true",
                  help="show context switches heat map (requires --task)"
                  )
parser.add_option("--kvm-exits",
                  dest="kvm_exits",
                  action="store_true",
                  help="show kvm exits heat map (requires --task)"
                  )
parser.add_option("--kvm-exit-types",
                  dest="kvm_exit_types",
                  action="store_true",
                  help="show kvm exit types bar charts (requires --task)"
                  )
parser.add_option("--label",
                  dest="label",
                  metavar="label",
                  help="label for the title (defaulst to the cdict file name)"
                  )
parser.add_option("-c", "--cap",
                  dest="cap_time",
                  help="(optional) cap the analysis to first <cap_time> msec"
                       " of capture (default=all)"
                  )
parser.add_option("-f", "--from",
                  dest="from_time",
                  help="(optional) start the analysis after first <from_time> msec"
                       " of capture (default=0)"
                  )
parser.add_option("--map",
                  dest="map",
                  action="store",
                  metavar="mapping csv file",
                  help="remap task names from mapping csv file"
                  )
parser.add_option("--convert",
                  dest="convert",
                  action="store",
                  metavar="new cdict file",
                  help="(Deprecated) migrate to new encoding with runtime aggregation into switch"
                  )
(options, args) = parser.parse_args()

if options.from_time:
    from_time = int(options.from_time) * 1000
if options.cap_time:
    # convert to usec
    cap_time = int(options.cap_time) * 1000 + from_time

if not args:
    print 'Missing cdict file'
    sys.exit(1)

cdict_file = args[0]
perf_dict = open_cdict(cdict_file, options.map)

df = DataFrame(perf_dict)
set_html_file(cdict_file)

# filter on usecs
if from_time:
    df = df[df['usecs'] >= from_time]
if cap_time:
    df = df[df['usecs'] <= cap_time]

if not options.label:
    options.label = os.path.splitext(os.path.basename(cdict_file))[0]

if options.convert:
    convert(df, options.convert)
    sys.exit(0)

if options.show_tids:
    res = df.groupby(['pid', 'task_name']).size()
    res.sort_values(ascending=False, inplace=True)
    print 'List of tids and task names sorted by context switches and kvm event count'
    print res
    sys.exit(0)

if options.successor_of_task:
    show_successors(df, options.successor_of_task, options.label)
    sys.exit(0)

# These options can be cumulative and all require a --task parameter to select tasks
if not options.task:
    print '--task <task_regex> is required'
    sys.exit(1)

if options.core_runtime:
    show_core_runs(df, options.task, options.label, True)

if options.core_switches:
    show_core_runs(df, options.task, options.label, False)

if options.core_loc:
    show_core_locality(df, options.task, options.label)

if options.switches or options.kvm_exits:
    show_sw_kvm_heatmap(df, options.task, options.label, options.switches, options.kvm_exits)

if options.kvm_exit_types:
    show_kvm_exit_types(df, options.task, options.label)



# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

# Module for credentials in Openstack
import os
import re


class Credentials(object):

    def get_credentials(self):
        dct = {}
        dct['username'] = self.rc_username
        dct['password'] = self.rc_password
        dct['auth_url'] = self.rc_auth_url
        dct['tenant_name'] = self.rc_tenant_name
        return dct

    def get_nova_credentials(self):
        dct = {}
        dct['username'] = self.rc_username
        dct['api_key'] = self.rc_password
        dct['auth_url'] = self.rc_auth_url
        dct['project_id'] = self.rc_tenant_name
        return dct

    def get_nova_credentials_v2(self):
        dct = self.get_nova_credentials()
        dct['version'] = 2
        return dct

    #
    # Read a openrc file and take care of the password
    # The 2 args are passed from the command line and can be None
    #
    def __init__(self, opts):
        self.rc_password = None
        self.rc_username = None
        self.rc_tenant_name = None
        self.rc_auth_url = None
        openrc_file = None
        pwd = None

        try:
            pwd = opts.password
        except AttributeError:
            pass
        try:
            openrc_file = opts.rc
        except AttributeError:
            pass

        if openrc_file:
            if os.path.exists(openrc_file):
                export_re = re.compile('export OS_([A-Z_]*)="?(.*)')
                for line in open(openrc_file):
                    line = line.strip()
                    mstr = export_re.match(line)
                    if mstr:
                        name = mstr.group(1)
                        value = mstr.group(2)
                        if value.endswith('"'):
                            value = value[:-1]
                        # get rid of password assignment
                        # echo "Please enter your OpenStack Password: "
                        # read -sr OS_PASSWORD_INPUT
                        # export OS_PASSWORD=$OS_PASSWORD_INPUT
                        if value.startswith('$'):
                            continue
                        # now match against wanted variable names
                        if name == 'USERNAME':
                            self.rc_username = value
                        elif name == 'AUTH_URL':
                            self.rc_auth_url = value
                        elif name == 'PASSWORD':
                            self.rc_password = value
                        elif name == 'TENANT_NAME':
                            self.rc_tenant_name = value
            else:
                raise ValueError('rc file does not exist %s' % (openrc_file))
        else:
            # no openrc file passed - we assume the variables have been
            # sourced by the calling shell
            # just check that they are present
            for varname in ['OS_USERNAME', 'OS_AUTH_URL', 'OS_TENANT_NAME']:
                if varname not in os.environ:
                    raise ValueError('%s is missing' % (varname))
            self.rc_username = os.environ['OS_USERNAME']
            self.rc_auth_url = os.environ['OS_AUTH_URL']
            self.rc_tenant_name = os.environ['OS_TENANT_NAME']

        # always override with CLI argument if provided
        if pwd:
            self.rc_password = pwd
        # if password not know, check from env variable
        elif self.rc_auth_url and not self.rc_password:
            if 'OS_PASSWORD' in os.environ:
                self.rc_password = os.environ['OS_PASSWORD']



#!/usr/bin/env python
# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
#
# Author: Alec Hothan
# -----------------------------------s----------------------
#
# A wrapper around the perf tool to capture various data related to context switches and
# KVM events
#
import os
import sys
from optparse import OptionParser
import re
import subprocess
import perf_formatter

perf_binary = 'perf'


# Curate the task names for a latency table
#  ---------------------------------------------------------------------------------------------------------------
#  Task                  |   Runtime ms  | Switches | Average delay ms | Maximum delay ms | Maximum delay at     |
# ---------------------------------------------------------------------------------------------------------------
#  qemu-system-x86:13568 |      5.762 ms |      102 | avg:    0.008 ms | max:    0.015 ms | max at: 9695849.502350 s
#  apache2:24581         |      0.397 ms |       10 | avg:    0.010 ms | max:    0.013 ms | max at: 9695849.359025 s
# The qemu task names need to be transformed into
# a type (emulator or vcpu0)
# By default, qemu-system-x86:13568 becomes qemu.vcpu0:13568 (for example)
#

def get_curated_latency_table(table):
    try:
        perf_formatter.init(opts)
    except ImportError:
        print 'Using default qemu task name mapping (no plugin found)'
    except ValueError:
        print 'Using default qemu task name mapping (OpenStack credentials not found, use -r or env variables)'

    lines = table.split('\n')
    results = []
    for line in lines:
        index = line.find('|')
        if index <= 0:
            results.append(line)
            continue
        tname = line[:index]
        trailer = line[index:]
        # count how many spaces at head
        non_space = 0
        len_tname = len(tname)
        while non_space < len_tname and tname[non_space] == ' ':
            non_space += 1
        tname = tname[non_space:]
        if tname.startswith('qemu-'):
            index = tname.rfind(':')
            if index > 0:
                tid = int(tname[index + 1:])
                header = ' ' * non_space
                tname = perf_formatter.get_task_name(tid, 'qemu') + ':' + str(tid)
                # if needed add more space to keep alignment of next column
                pad_len = len_tname - non_space - len(tname)
                if pad_len > 0:
                    tname += ' ' * pad_len
                line = header + tname + trailer
        results.append(line)
    return '\n'.join(results)

def perf_record(opts, cs=True, kvm=True):
    perf_cmd = [perf_binary, 'record', '-a']
    if cs:
        perf_cmd += ['-e', 'sched:*']
    if kvm:
        perf_cmd += ['-e', 'kvm:*']
    perf_cmd += ['sleep', str(opts.seconds)]
    print 'Recording with: ' + ' '.join(perf_cmd)
    rc = subprocess.call(perf_cmd)
    if rc:
        print 'Error recording traces'
        print 'You might need to run this script as root or with sudo'
        return False
    return True


def capture_stats(opts, stats_filename):
    # perf sched latency -s switch
    cmd = [perf_binary, 'sched', 'latency', '-s', 'switch']
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=None)
    results, errors = process.communicate()
    if errors:
        print 'Error displaying scheduling latency'
    else:
        # curate the process names before displaying
        results = get_curated_latency_table(results)

    with open(stats_filename, 'w') as ff:
        ff.write(results)

        # need to let the results to be flushed out
        ff.flush()

        rc = subprocess.call([perf_binary, 'kvm', 'stat', 'report'], stdout=ff, stderr=ff)
        if rc:
            print 'Error displaying kvm stats'
        else:
            print 'Stats captured in ' + stats_filename
    os.chmod(stats_filename, 0664)


def capture(opts, run_name):

    # If this is set we skip the capture
    perf_data_filename = opts.perf_data
    if perf_data_filename:
        print 'Skipping capture, using ' + perf_data_filename
    else:
        # need to capture traces
        print 'Capturing perf data for %d seconds...' % (opts.seconds)
        if not perf_record(opts):
            return
        perf_data_filename = 'perf.data'
        print 'Traces captured in perf.data'

    # collect stats
    if opts.all or opts.stats:
        if opts.perf_data:
            print 'Stats capture from provided perf data file not supported'
        else:
            stats_filename = opts.dest_folder + run_name + ".stats"
            capture_stats(opts, stats_filename)

    # create cdict from the perf data file

    if opts.all or opts.switches:
        try:
            cdict_filename = opts.dest_folder + run_name + '.cdict'
            # try to run this script through the perf tool itself as it is faster
            rc = subprocess.call([perf_binary, 'script', '-s', 'mkcdict_perf_script.py', '-i', perf_data_filename])
            if rc == 255:
                print '   ERROR: perf is not built with the python scripting extension - aborting...'
            else:
                # success result is in perf.cdict, so need to rename it
                os.rename('perf.cdict', cdict_filename)
                os.chmod(cdict_filename, 0664)
                print 'Created file: ' + cdict_filename
        except OSError:
            print 'Error: perf does not seems to be installed'

if __name__ == '__main__':
    parser = OptionParser(usage="usage: %prog [options] [<run-name>]")

    parser.add_option('--stats', dest='stats',
                      action='store_true',
                      default=False,
                      help='capture context switches and kvm exit stats to <run-name>.stats')

    parser.add_option('--switches', dest='switches',
                      action='store_true',
                      default=False,
                      help='capture detailed context switch and kvm traces and create <run-name>.cdict if not none')

    parser.add_option('--all', dest='all',
                      action='store_true',
                      default=False,
                      help='capture all traces and stats into <run-name>.cdict and <run-name>.stats')

    parser.add_option('-s', '--sleep', dest='seconds',
                      action='store',
                      default=1,
                      type="int",
                      help='capture duration in seconds, defaults to 1 second',
                      metavar='<seconds>')

    parser.add_option('--dest-folder', dest='dest_folder',
                      action='store',
                      default='./',
                      help='destination folder where to store results (default: current folder)',
                      metavar='<dest folder>')

    parser.add_option('--use-perf-data', dest='perf_data',
                      action='store',
                      help='use given perf data file (do not capture)',
                      metavar='<perf data file>')

    parser.add_option("--map",
                      dest="map",
                      action="store",
                      metavar="mapping csv file",
                      help="remap task names from mapping csv file"
                      )

    parser.add_option('--use-perf', dest='perf',
                      action='store',
                      help='use given perf binary',
                      metavar='<perf binary>')

    parser.add_option("--remap",
                      dest="remap",
                      action="store",
                      metavar="cdict_file",
                      help="remap the passed cdict file using the given map file (see --map)"
                      )

    parser.add_option('-r', '--rc', dest='rc',
                      action='store',
                      help='source OpenStack credentials from rc file',
                      metavar='<openrc_file>')

    parser.add_option('-p', '--password', dest='passwordd',
                      action='store',
                      help='OpenStack password',
                      metavar='<password>')

    (opts, args) = parser.parse_args()

    if len(args) > 1:
        print 'This script requires 1 argument for the run name (any valid file name without extension)'
        sys.exit(0)
    if len(args):
        run_name = args[0]
    else:
        run_name = 'perf'

    if not os.path.isdir(opts.dest_folder):
        print 'Invalid destination folder: ' + opts.dest_folder
        sys.exit(2)
    if not opts.dest_folder.endswith('/'):
        opts.dest_folder += '/'

    # chek the map file is valid if provided
    if opts.map:
        if not os.path.exists(opts.map):
            print 'ERROR: Invalid mapping file: ' + opts.map
            sys.exit(1)

    # pick at least one command
    if opts.remap:
        if not opts.map:
            print 'ERROR: remap command requires a csv mapping file (--map)'
            sys.exit(1)
        perf_dict = perf_formatter.open_cdict(opts.remap, opts.map)
        perf_formatter.write_cdict(opts.remap, perf_dict)
        sys.exit(0)

    if not (opts.all | opts.switches | opts.stats):
        print 'Pick at least one of --stats, --switches, --all'
        sys.exit(3)

    CFG_FILE = '.mkcdict.cfg'

    # check if there is a config file in the current directory
    if os.path.isfile(CFG_FILE):
        # load options from the config file if they are set to None
        print 'Loading options from ' + CFG_FILE
        opt_re = re.compile(' *(\w*) *[=:]+ *([\w/\-\.]*)')
        with open(CFG_FILE, 'r') as ff:
            for line in ff:
                m = opt_re.match(line)
                if m:
                    cfg_name = m.group(1)
                    try:
                        if getattr(opts, cfg_name) is None:
                            setattr(opts, cfg_name, m.group(2))
                    except AttributeError:
                        pass

    if opts.perf_data and not os.path.isfile(opts.perf_data):
        print 'Cannot find perf data file: ' + opts.perf_data
        sys.exit(1)

    if opts.perf:
        perf_binary = opts.perf
        print 'Overriding perf binary with: ' + perf_binary

    capture(opts, run_name)



#
# See the perf-trace-python Documentation for the list of available functions.
# perf python script for creating the cdict files out of a perf.data file
# To be used by the perf-sched post processing script.
#
# This script reads a perf binary file (through perf script -s) and generates a cdict file
# named perf.cdict.
# The cdict file is a compressed json dump of a python dictionary that contains
# a subset of the perf traces in a form that is ready to be loaded into a pandas dataframe.
#
# Functions in this script are also called from mkcdict.py when the python scripting of perf is not compiled in.
#
import os
import sys
from os.path import expanduser
import re

# Location of the perf python helper files
try:
    sys.path.append(os.environ['PERF_EXEC_PATH'] +
                    '/scripts/python/perf-script-Util/lib/Perf/Trace')
except KeyError:
    pass
sys.path.append(expanduser('~/perf-trace-lib'))

try:
    from perf_trace_context import *
    from Core import *
except ImportError:
    pass

try:
    # try to use the faster version if available
    from msgpack import packb
except ImportError:
    # else fall back to the pure python version (slower)
    from umsgpack import packb
import zlib

event_name_list = []
cpu_list = []
usecs_list = []
pid_list = []
comm_list = []
duration_list = []
next_pid_list = []
next_comm_list = []

# a dict of task names indexed by tid
name_by_tid = {}

# A dict of counts indexed by event name
# counts how many are being ignored (not counted) in the cdict
event_drops = {}

# A dict of counts indexed by event name
# counts how many are being counted and added to the cdict
event_counts = {}

def drop_event(event_name):
    try:
        event_drops[event_name] += 1
    except KeyError:
        event_drops[event_name] = 1

def count_event(event_name):
    try:
        event_counts[event_name] += 1
    except KeyError:
        event_counts[event_name] = 1

def trace_begin():
    global plugin_convert_name

    # try to import
    try:
        from mkcdict_plugin import plugin_init
        print 'Initializing plugin...'
        if plugin_init():
            from mkcdict_plugin import plugin_convert_name
    except ImportError:
        plugin_convert_name = None

def trace_end():
    # report dropped kvm events
    print 'Dropped events (not stored in cdict file):'
    for name in sorted(event_drops, key=event_drops.get, reverse=True):
        print '   %6d %s' % (event_drops[name], name)
    print
    print 'Events stored in cdict file:'
    for name in sorted(event_counts, key=event_counts.get, reverse=True):
        print '   %6d %s' % (event_counts[name], name)
    print
    # build cdict
    res = {'event': event_name_list,
           'cpu': cpu_list,
           'usecs': usecs_list,
           'pid': pid_list,
           'task_name': comm_list,
           'duration': duration_list,
           'next_pid': next_pid_list,
           'next_comm': next_comm_list}
    print 'End of trace, marshaling and compressing...'
    compressed = zlib.compress(packb(res))
    with open('perf.cdict', 'w') as ff:
        ff.write(compressed)
    print 'Compressed dictionary written to perf.cdict %d entries size=%d bytes' % \
          (len(cpu_list), len(compressed))

uuid_re = re.compile('-uuid ([a-fA-F0-9\-]*)')
# /proc/pid/cpuset output
# /machine/instance-000065d7.libvirt-qemu/emulator
# /machine/instance-000065d3.libvirt-qemu/vcpu0
cpuset_re = re.compile('/machine/(instance-[a-fA-F0-9]*).libvirt-qemu/(\w*)')

def decode_cpuset(pid):
    name = None
    thread_type = None
    try:
        with open('/proc/%d/cpuset' % (pid)) as f:
            cpuset = f.read()
            res = cpuset_re.match(cpuset)
            if res:
                name = res.group(1)
                thread_type = res.group(2)
    except IOError:
        pass
    return name, thread_type

def decode_pid(pid):
    name = None
    uuid = None
    thread_type = None
    try:
        with open('/proc/%d/cmdline' % (pid)) as f:
            cmdline = f.read()
            # this is a nul separated tokens
            cmdline = cmdline.replace('\x00', ' ')
            if cmdline.startswith('/usr/bin/qemu-system'):
                res = uuid_re.search(cmdline)
                if res:
                    uuid = res.group(1)
                    name, thread_type = decode_cpuset(pid)
    except IOError:
        pass
    return name, uuid, thread_type

def get_final_name(tid, name):
    if not tid:
        return name
    try:
        return name_by_tid[tid]
    except KeyError:
        pass
    # check if it is a kvm thread from the look of the name
    libvirt_name, uuid, thread_type = decode_pid(tid)
    if libvirt_name:
        if plugin_convert_name:
            name = plugin_convert_name(name, tid, libvirt_name, uuid, thread_type)
        # append the thread type to the name
        name += '.' + thread_type

    name_by_tid[tid] = name
    return name

def get_usecs(secs, nsecs):
    global epoch
    try:
        usecs = (secs * 1000000) + (nsecs / 1000) - epoch
    except NameError:
        epoch = (secs * 1000000) + (nsecs / 1000)
        usecs = 0
    return usecs

def add_event(name, cpu, secs, nsecs, pid, comm, duration=0, next_pid=0, next_comm=None):
    event_name_list.append(name)
    cpu_list.append(cpu)
    usecs_list.append(get_usecs(secs, nsecs))
    pid_list.append(pid)
    comm_list.append(get_final_name(pid, comm))
    # duration in usec
    duration_list.append(duration / 1000)
    next_pid_list.append(next_pid)
    next_comm_list.append(get_final_name(next_pid, next_comm))
    count_event(name)

def add_kvm_event(name, cpu, secs, nsecs, pid, comm, prev_usecs, reason=None):
    usecs = get_usecs(secs, nsecs)
    event_name_list.append(name)
    cpu_list.append(cpu)
    usecs_list.append(usecs)
    pid_list.append(pid)
    comm_list.append(get_final_name(pid, comm))
    duration_list.append(usecs - prev_usecs)
    next_pid_list.append(None)
    next_comm_list.append(reason)
    count_event(name)
    return usecs

def sched__sched_stat_sleep(event_name, context, common_cpu,
                            common_secs, common_nsecs, common_pid, common_comm,
                            comm, pid, delay):
    # the delay (time slept) applies to comm/pid
    # not common_comm/common_pid
    add_event(event_name, common_cpu, common_secs, common_nsecs, pid, comm, delay)


def sched__sched_wakeup_new(event_name, context, common_cpu,
                            common_secs, common_nsecs, common_pid, common_comm,
                            comm, pid, prio, success,
                            target_cpu):
    drop_event(event_name)


def sched__sched_wakeup(event_name, context, common_cpu,
                        common_secs, common_nsecs, common_pid, common_comm,
                        comm, pid, prio, success,
                        target_cpu):
    drop_event(event_name)

# A dict of runtime delays accumulated indexed by cpu
runtime_by_cpu = {}

def sched__sched_stat_runtime(event_name, context, common_cpu,
                              common_secs, common_nsecs, common_pid, common_comm,
                              comm, pid, runtime, vruntime):
    try:
        runtime_by_cpu[common_cpu] += runtime
    except KeyError:
        # the counter is set in on the first sched switch for each cpu
        pass

def sched__sched_switch(event_name, context, common_cpu,
                        common_secs, common_nsecs, common_pid, common_comm,
                        prev_comm, prev_pid, prev_prio, prev_state,
                        next_comm, next_pid, next_prio):
    try:
        runtime = runtime_by_cpu[common_cpu]
        add_event(event_name, common_cpu, common_secs, common_nsecs, prev_pid, prev_comm, runtime, next_pid, next_comm)
    except KeyError:
        pass
    runtime_by_cpu[common_cpu] = 0

def sched__sched_stat_iowait(event_name, context, common_cpu,
                             common_secs, common_nsecs, common_pid, common_comm,
                             comm, pid, delay):
    add_event(event_name, common_cpu, common_secs, common_nsecs, pid, comm, delay)


class KvmTime(object):
    def __init__(self, tid):
        self.tid = tid
        # all units are usecs since beginning of traces
        self.entry_time = 0
        self.exit_time = 0

    def add_entry(self, cpu, secs, nsecs, comm):
        if self.exit_time:
            self.entry_time = add_kvm_event('kvm_entry', cpu, secs, nsecs, self.tid,
                                            comm, self.exit_time)
        else:
            self.entry_time = get_usecs(secs, nsecs)

    def add_exit(self, cpu, secs, nsecs, comm, reason):
        if self.entry_time:
            self.exit_time = add_kvm_event('kvm_exit', cpu, secs, nsecs, self.tid,
                                           comm, self.entry_time, reason)
        else:
            self.exit_time = get_usecs(secs, nsecs)

# A dict of last kvm entry/exit time indexed by tid
kvm_time_dict = {}

def kvm__kvm_entry(event_name, context, common_cpu,
                   common_secs, common_nsecs, common_pid, common_comm,
                   vcpu_id):
    try:
        kt = kvm_time_dict[common_pid]
    except:
        kt = KvmTime(common_pid)
        kvm_time_dict[common_pid] = kt
    kt.add_entry(common_cpu, common_secs, common_nsecs, common_comm)

def kvm__kvm_exit(event_name, context, common_cpu,
                  common_secs, common_nsecs, common_pid, common_comm,
                  exit_reason, guest_rip, isa, info1,
                  info2):
    try:
        kt = kvm_time_dict[common_pid]
    except:
        kt = KvmTime(common_pid)
        kvm_time_dict[common_pid] = kt
    kt.add_exit(common_cpu, common_secs, common_nsecs, common_comm, exit_reason)

# These are scale down versions of kvm__kvm_entry/exit that only require the minimum arguments
# used for manual parsing when the perf python extension is not compiled in
def kvm_entry(common_cpu, common_secs, common_nsecs, common_pid, common_comm):
    try:
        kt = kvm_time_dict[common_pid]
    except:
        kt = KvmTime(common_pid)
        kvm_time_dict[common_pid] = kt
    kt.add_entry(common_cpu, common_secs, common_nsecs, common_comm)

def kvm_exit(common_cpu, common_secs, common_nsecs, common_pid, common_comm, exit_reason):
    try:
        kt = kvm_time_dict[common_pid]
    except:
        kt = KvmTime(common_pid)
        kvm_time_dict[common_pid] = kt
    kt.add_exit(common_cpu, common_secs, common_nsecs, common_comm, exit_reason)

def sched__sched_process_hang(event_name, context, common_cpu,
                              common_secs, common_nsecs, common_pid, common_comm,
                              comm, pid):
    drop_event(event_name)


def sched__sched_pi_setprio(event_name, context, common_cpu,
                            common_secs, common_nsecs, common_pid, common_comm,
                            comm, pid, oldprio, newprio):
    drop_event(event_name)


def sched__sched_stat_blocked(event_name, context, common_cpu,
                              common_secs, common_nsecs, common_pid, common_comm,
                              comm, pid, delay):
    drop_event(event_name)


def sched__sched_stat_wait(event_name, context, common_cpu,
                           common_secs, common_nsecs, common_pid, common_comm,
                           comm, pid, delay):
    drop_event(event_name)


def sched__sched_process_exec(event_name, context, common_cpu,
                              common_secs, common_nsecs, common_pid, common_comm,
                              filename, pid, old_pid):
    drop_event(event_name)


def sched__sched_process_fork(event_name, context, common_cpu,
                              common_secs, common_nsecs, common_pid, common_comm,
                              parent_comm, parent_pid, child_comm, child_pid):
    drop_event(event_name)


def sched__sched_process_wait(event_name, context, common_cpu,
                              common_secs, common_nsecs, common_pid, common_comm,
                              comm, pid, prio):
    drop_event(event_name)


def sched__sched_wait_task(event_name, context, common_cpu,
                           common_secs, common_nsecs, common_pid, common_comm,
                           comm, pid, prio):
    drop_event(event_name)


def sched__sched_process_exit(event_name, context, common_cpu,
                              common_secs, common_nsecs, common_pid, common_comm,
                              comm, pid, prio):
    drop_event(event_name)


def sched__sched_process_free(event_name, context, common_cpu,
                              common_secs, common_nsecs, common_pid, common_comm,
                              comm, pid, prio):
    drop_event(event_name)


def sched__sched_migrate_task(event_name, context, common_cpu,
                              common_secs, common_nsecs, common_pid, common_comm,
                              comm, pid, prio, orig_cpu,
                              dest_cpu):
    drop_event(event_name)


def sched__sched_kthread_stop_ret(event_name, context, common_cpu,
                                  common_secs, common_nsecs, common_pid, common_comm,
                                  ret):
    drop_event(event_name)


def sched__sched_kthread_stop(event_name, context, common_cpu,
                              common_secs, common_nsecs, common_pid, common_comm,
                              comm, pid):
    drop_event(event_name)


def kvm__kvm_async_pf_completed(event_name, context, common_cpu,
                                common_secs, common_nsecs, common_pid, common_comm,
                                address, gva):
    drop_event(event_name)


def kvm__kvm_async_pf_ready(event_name, context, common_cpu,
                            common_secs, common_nsecs, common_pid, common_comm,
                            token, gva):
    drop_event(event_name)


def kvm__kvm_async_pf_not_present(event_name, context, common_cpu,
                                  common_secs, common_nsecs, common_pid, common_comm,
                                  token, gva):
    drop_event(event_name)


def kvm__kvm_async_pf_doublefault(event_name, context, common_cpu,
                                  common_secs, common_nsecs, common_pid, common_comm,
                                  gva, gfn):
    drop_event(event_name)


def kvm__kvm_try_async_get_page(event_name, context, common_cpu,
                                common_secs, common_nsecs, common_pid, common_comm,
                                gva, gfn):
    drop_event(event_name)


def kvm__kvm_age_page(event_name, context, common_cpu,
                      common_secs, common_nsecs, common_pid, common_comm,
                      hva, gfn, referenced):
    drop_event(event_name)


def kvm__kvm_fpu(event_name, context, common_cpu,
                 common_secs, common_nsecs, common_pid, common_comm,
                 load):
    drop_event(event_name)


def kvm__kvm_mmio(event_name, context, common_cpu,
                  common_secs, common_nsecs, common_pid, common_comm,
                  type, len, gpa, val):
    drop_event(event_name)


def kvm__kvm_ack_irq(event_name, context, common_cpu,
                     common_secs, common_nsecs, common_pid, common_comm,
                     irqchip, pin):
    drop_event(event_name)


def kvm__kvm_msi_set_irq(event_name, context, common_cpu,
                         common_secs, common_nsecs, common_pid, common_comm,
                         address, data):
    drop_event(event_name)


def kvm__kvm_ioapic_set_irq(event_name, context, common_cpu,
                            common_secs, common_nsecs, common_pid, common_comm,
                            e, pin, coalesced):
    drop_event(event_name)


def kvm__kvm_set_irq(event_name, context, common_cpu,
                     common_secs, common_nsecs, common_pid, common_comm,
                     gsi, level, irq_source_id):
    drop_event(event_name)


def kvm__kvm_userspace_exit(event_name, context, common_cpu,
                            common_secs, common_nsecs, common_pid, common_comm,
                            reason, errno):
    drop_event(event_name)


def kvm__kvm_track_tsc(event_name, context, common_cpu,
                       common_secs, common_nsecs, common_pid, common_comm,
                       vcpu_id, nr_vcpus_matched_tsc, online_vcpus, use_master_clock,
                       host_clock):
    drop_event(event_name)


def kvm__kvm_update_master_clock(event_name, context, common_cpu,
                                 common_secs, common_nsecs, common_pid, common_comm,
                                 use_master_clock, host_clock, offset_matched):
    drop_event(event_name)


def kvm__kvm_write_tsc_offset(event_name, context, common_cpu,
                              common_secs, common_nsecs, common_pid, common_comm,
                              vcpu_id, previous_tsc_offset, next_tsc_offset):
    drop_event(event_name)


def kvm__vcpu_match_mmio(event_name, context, common_cpu,
                         common_secs, common_nsecs, common_pid, common_comm,
                         gva, gpa, write, gpa_match):
    drop_event(event_name)


def kvm__kvm_emulate_insn(event_name, context, common_cpu,
                          common_secs, common_nsecs, common_pid, common_comm,
                          rip, csbase, len, insn,
                          flags, failed):
    drop_event(event_name)


def kvm__kvm_skinit(event_name, context, common_cpu,
                    common_secs, common_nsecs, common_pid, common_comm,
                    rip, slb):
    drop_event(event_name)


def kvm__kvm_invlpga(event_name, context, common_cpu,
                     common_secs, common_nsecs, common_pid, common_comm,
                     rip, asid, address):
    drop_event(event_name)


def kvm__kvm_nested_intr_vmexit(event_name, context, common_cpu,
                                common_secs, common_nsecs, common_pid, common_comm,
                                rip):
    drop_event(event_name)


def kvm__kvm_nested_vmexit_inject(event_name, context, common_cpu,
                                  common_secs, common_nsecs, common_pid, common_comm,
                                  exit_code, exit_info1, exit_info2, exit_int_info,
                                  exit_int_info_err, isa):
    drop_event(event_name)


def kvm__kvm_nested_vmexit(event_name, context, common_cpu,
                           common_secs, common_nsecs, common_pid, common_comm,
                           rip, exit_code, exit_info1, exit_info2,
                           exit_int_info, exit_int_info_err, isa):
    drop_event(event_name)


def kvm__kvm_nested_intercepts(event_name, context, common_cpu,
                               common_secs, common_nsecs, common_pid, common_comm,
                               cr_read, cr_write, exceptions, intercept):
    drop_event(event_name)


def kvm__kvm_nested_vmrun(event_name, context, common_cpu,
                          common_secs, common_nsecs, common_pid, common_comm,
                          rip, vmcb, nested_rip, int_ctl,
                          event_inj, npt):
    drop_event(event_name)


def kvm__kvm_pv_eoi(event_name, context, common_cpu,
                    common_secs, common_nsecs, common_pid, common_comm,
                    apicid, vector):
    drop_event(event_name)


def kvm__kvm_eoi(event_name, context, common_cpu,
                 common_secs, common_nsecs, common_pid, common_comm,
                 apicid, vector):
    drop_event(event_name)


def kvm__kvm_apic_accept_irq(event_name, context, common_cpu,
                             common_secs, common_nsecs, common_pid, common_comm,
                             apicid, dm, tm, vec,
                             coalesced):
    drop_event(event_name)


def kvm__kvm_apic_ipi(event_name, context, common_cpu,
                      common_secs, common_nsecs, common_pid, common_comm,
                      icr_low, dest_id):
    drop_event(event_name)


def kvm__kvm_pic_set_irq(event_name, context, common_cpu,
                         common_secs, common_nsecs, common_pid, common_comm,
                         chip, pin, elcr, imr,
                         coalesced):
    drop_event(event_name)


def kvm__kvm_apic(event_name, context, common_cpu,
                  common_secs, common_nsecs, common_pid, common_comm,
                  rw, reg, val):
    drop_event(event_name)


def kvm__kvm_cr(event_name, context, common_cpu,
                common_secs, common_nsecs, common_pid, common_comm,
                rw, cr, val):
    drop_event(event_name)


def kvm__kvm_msr(event_name, context, common_cpu,
                 common_secs, common_nsecs, common_pid, common_comm,
                 write, ecx, data, exception):
    drop_event(event_name)


def kvm__kvm_page_fault(event_name, context, common_cpu,
                        common_secs, common_nsecs, common_pid, common_comm,
                        fault_address, error_code):
    drop_event(event_name)


def kvm__kvm_inj_exception(event_name, context, common_cpu,
                           common_secs, common_nsecs, common_pid, common_comm,
                           exception, has_error, error_code):
    drop_event(event_name)


def kvm__kvm_inj_virq(event_name, context, common_cpu,
                      common_secs, common_nsecs, common_pid, common_comm,
                      irq):
    drop_event(event_name)


def kvm__kvm_cpuid(event_name, context, common_cpu,
                   common_secs, common_nsecs, common_pid, common_comm,
                   function, rax, rbx, rcx,
                   rdx):
    drop_event(event_name)


def kvm__kvm_pio(event_name, context, common_cpu,
                 common_secs, common_nsecs, common_pid, common_comm,
                 rw, port, size, count, val=None):
    drop_event(event_name)


def kvm__kvm_hv_hypercall(event_name, context, common_cpu,
                          common_secs, common_nsecs, common_pid, common_comm,
                          rep_cnt, rep_idx, ingpa, outgpa,
                          code, fast):
    drop_event(event_name)


def kvm__kvm_hypercall(event_name, context, common_cpu,
                       common_secs, common_nsecs, common_pid, common_comm,
                       nr, a0, a1, a2,
                       a3):
    drop_event(event_name)

def kvm__kvm_ple_window(event_name, context, common_cpu,
                        common_secs, common_nsecs, common_pid, common_comm,
                        grow, vcpu_id, new, old):
    drop_event(event_name)

def trace_unhandled(event_name, context, event_fields_dict):
    print 'unhandled ' + event_name
    print ' '.join(['%s=%s' % (k, str(v)) for k, v in sorted(event_fields_dict.items())])
